#ifndef FEPKANAENGINE_H
#define FEPKANAENGINE_H

// INCLUDES
#include <e32base.h>
#include <e32std.h>

#define BUFSIZE 2048

#define COMMIT_COMMIT 1
#define COMMIT_REG_KANJI 2
#define COMMIT_REG_YOMI 3

#define KAKUTEI_ENTER 0
#define KAKUTEI_MAJOR 1

#define NO_SET 0
#define SET_POST 1

#define CAPS_HANKAKU 1
#define CAPS_HIRAGANA 2
#define CAPS_ZENKATA 3

// Forward declarations
class CSKKmControl;
class CFepIndicator;

class DicPtr {
public:
  DicPtr();
  void NewContents(RBuf&);
  void NewContents(TDesC16&);
  void UpdateContents(RBuf&);
  void SetNext(DicPtr*);
  RBuf contents;
  DicPtr *next;
private:
};

class RegEntry {
 public:
    RBuf yomi,kanji,regbuf, label;
    DicPtr *entry;
    RegEntry *prev;
    TUint post_char;

 private:
};

class CSKKmKanaEngine : public CBase
{
 public:
    CSKKmKanaEngine(CSKKmControl& aControl, CFepIndicator* aIndicator);
    ~CSKKmKanaEngine();
    // Static construction
    static CSKKmKanaEngine* NewL (CSKKmControl& aControl, CFepIndicator* aIndicator);

    // Second phase construction
    void ConstructL();

    TPtrC AppendCharacter(TUint , TUint* , TUint* , TUint);
    TPtrC AppendAbbrev(TUint , TUint*);
    TPtrC DeleteCharacter(void);
    void ClearBuffer(void);
    void PrepareRetbuf();
    TPtrC Buffer(void);
    TPtrC Retbuf(void);
    TPtrC Henkan(TUint *, TUint *);
    TPtrC henkan_backward(TUint *);
    TPtrC Kakutei(TUint *, TUint*, TUint);
    TPtrC backspace(TUint *, TUint*, TUint*);
    TPtrC remove_entry(TUint *);
    void KanjiIn();
    void SetKakutei();
    void ClearKakutei();
    void toggle_katakana(TUint);
    void toggle_hankaku(TUint);
    void set_hiragana();
    void set_katakana();
    void set_hankaku();
    void save_dic();

 private:
    TUint SearchKanji(int);
    void load_dic();
    DicPtr *search_entry();
    int candidate(DicPtr*, int);
    void set_frequent(DicPtr*, int);
    void start_reg_kanji();
    void commit_reg_kanji();
    void cancel_reg_kanji();
    void append_kanji(RegEntry*);
    void create_new_kanji_entry(RegEntry*);
    void ClearRegbuf();
    void ClearRetbuf();
    void remove_prev_strs();
    void show_indicator();

 public:
    TUint kanji_mode;

 private:
    CSKKmControl&	iFepControl;	// not owned
    CFepIndicator* iIndicator;
    TBuf<BUFSIZE> iBuffer;

    TBuf<10> mark_registering,
	mark_in_conversion_reg,mark_in_conversion,
	mark_start_conversion_reg, mark_start_conversion;

    char prev_strs[BUFSIZE];
    int n_conv;
    int henkan_candidate;
    int c_begin, c_end;
    TBuf<BUFSIZE> yomi;
    TBuf<BUFSIZE> kanji;
    TBuf<BUFSIZE> kakutei;
    TBuf<BUFSIZE> retbuf;
    TBuf<BUFSIZE> regbuf;
    DicPtr *first, *found_entry;
    RegEntry *cur_reg_entry;
    TUint post_char;
    TUint katakana;
    TUint hankaku;
    char dicdrive;
    TBuf<BUFSIZE> ucsbuf, ucsbuf2;
    TBuf8<BUFSIZE> utfbuf;
};

#endif // FEPKANAENGINE_H

// end of file
