#include <eikappui.h>
#include <eikenv.h>         //CEikEnv
#include <bautils.h>	// for BaflUtils
#include <aknedsts.h> // for CAknEdwinState
#include "skkm.h"
#include "fepcontrol.h"
#include "kanaengine.h"
#include "fepindicator.h"

#define KEY_CHANGE_ALPHA 163 // 
// #define KEY_CHANGE_ALPHA2 10 // C-j
#define KEY_CHANGE_ALPHA2 15 // C-o

#define KEY_SAVE_DIC 8364 // 
#define KEY_SAVE_DIC2 23 // C-w

#define KEY_END_KANA 'l'
#define KEY_REMOVE_ENTRY 'X'
#define KEY_HENKAN_BACKWARD 'x'
#define KEY_START_ABBREV '/'
#define KEY_TOGGLE_KATAKANA 'q'
#define KEY_TOGGLE_HANKAKU 'Q'

/*
163 // 
174 // trade mark (R)
191 // reverse question mark (?)
8364 // 
*/

//_LIT(KT9FepResourceFileName, "z:\\system\\fep\\aknfep.rsc");

/////////////////////////////////////////////////////////////////////////////////
//
//Class CSKKmControl
//
/////////////////////////////////////////////////////////////////////////////////
/**
 * Two phase construction
 */
CSKKmControl* CSKKmControl::NewL(CSKKm& aFep)
	{
	CSKKmControl* self=new(ELeave) CSKKmControl(aFep);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop(); //self
	return self;
	}

CSKKmControl::CSKKmControl(CSKKm& aFep)
  :iFep(aFep), 
   iInputCapabilities(TCoeInputCapabilities::ENone),
   iInputMode(EAknEditorNullInputMode),
   iInsideInlineEditingTransaction(EFalse),
   kana_mode(1),
   abbrev_mode(0),
   capabilities(0)
{
}

void CSKKmControl::ConstructL()
	{
	// Set up fep key catching control - front window, null size, non-focusing
	CreateWindowL();
	SetNonFocusing();
	
	RWindow& window = Window();
	window.SetOrdinalPosition(0, ECoeWinPriorityFep);
	TPoint fepControlPos(0, 0);

	SetExtent(fepControlPos, TSize(0,0));
	window.SetExtent(fepControlPos, TSize(0,0));
	window.SetNonFading(ETrue);

	// Create status pane indicator
	iIndicator = CFepIndicator::NewL();

	// Create kana input engine
	iKanaEngine = CSKKmKanaEngine::NewL(*this, iIndicator);

	// The reason why we need to set up the T9 resource is that, Series 60 platform integrates the 
	// T9 engine into the system, and some on-board applications expect this resource to be
	// present, if not, those applications will not work. The T9 resource is loaded when the T9 FEP
	// start up.
	//
	// As a workaround solution, we load the T9 resource in our own FEP. 
/*
	TFileName resourceName(KT9FepResourceFileName);
	BaflUtils::NearestLanguageFile(iEikonEnv->FsSession(), resourceName);
	iResId = iEikonEnv->AddResourceFileL(resourceName);
*/
	((CCoeAppUi*)iEikonEnv->AppUi())->AddToStackL(this, 
		ECoeStackPriorityFep, ECoeStackFlagSharable|ECoeStackFlagRefusesFocus);
	}			

CSKKmControl::~CSKKmControl()
    {
	if (iKanaEngine)
		{
		delete iKanaEngine;
		iKanaEngine = NULL;
		}
	
	if (iIndicator)
		{
		delete iIndicator;
		iIndicator = NULL;
		}

//	iEikonEnv->DeleteResourceFile(iResId);

	((CCoeAppUi*)iEikonEnv->AppUi())->RemoveFromStack(this);
	}

// Please consult SDK for more information on Inline Editing transaction
void CSKKmControl::StartInlineEditL(MCoeFepAwareTextEditor& aFepAwareTextEditor, const TDesC& aInitialInlineText, TInt aPositionOfInsertionPointInInlineText, TBool aCursorVisibility)
	{
	aFepAwareTextEditor.StartFepInlineEditL(aInitialInlineText, aPositionOfInsertionPointInInlineText, 
			aCursorVisibility, NULL, *this, *this);

	iInsideInlineEditingTransaction = ETrue;
	}

// Please consult SDK for more information on Inline Editing transaction
void CSKKmControl::UpdateInlineEditL(const TDesC& aNewInlineText, TInt aPositionOfInsertionPointInInlineText)
	{
	iInputCapabilities.FepAwareTextEditor()->UpdateFepInlineTextL(aNewInlineText, aPositionOfInsertionPointInInlineText);						
	}

// Please consult SDK for more information on Inline Editing transaction
void CSKKmControl::CancelInlineEdit(MCoeFepAwareTextEditor& aFepAwareTextEditor)
{
  aFepAwareTextEditor.CancelFepInlineEdit();
  iInsideInlineEditingTransaction = EFalse;

  ClearBuffer();
  ClearKakutei();
}

// Please consult SDK for more information on Inline Editing transaction
void CSKKmControl::CommitInlineEditL(MCoeFepAwareTextEditor& aFepAwareTextEditor, TBool aCursorVisibility)
{
  aFepAwareTextEditor.SetInlineEditingCursorVisibilityL(aCursorVisibility);
  aFepAwareTextEditor.CommitFepInlineEditL(*iCoeEnv);
  iInsideInlineEditingTransaction = EFalse;

  ClearBuffer();
  ClearKakutei();
}


void CSKKmControl::GetFormatOfFepInlineText(TCharFormat& /*aFormat*/, TInt& /*aNumberOfCharactersWithSameFormat*/, TInt /*aPositionOfCharacter*/) const
	{
	// No implementation
	}


void CSKKmControl::HandlePointerEventInInlineTextL(TPointerEvent::TType /*aType*/, TUint /*aModifiers*/, TInt /*aPositionInInlineText*/)
	{
	// not supported in series 60 platform 1.x
	}

TKeyResponse CSKKmControl::OfferKeyEventL(const TKeyEvent& aKeyEvent, TEventCode aEventCode)
	{
	TInt keyCode = aKeyEvent.iCode;
	if (aEventCode != EEventKey)
		{
		return EKeyWasNotConsumed;
		}

	if (((keyCode == EKeyCBA1) || (keyCode == EKeyCBA2)) || (keyCode == EKeyMenu))
		{
		return EKeyWasNotConsumed;
		}

	if (iFep.IsTurnedOnByL(aKeyEvent) || iFep.IsTurnedOffByL(aKeyEvent))
		{
		return EKeyWasConsumed;
		}

/*
	if (!iFep.IsOn())
		return EKeyWasNotConsumed;
*/

	TKeyResponse response = EKeyWasNotConsumed;
	switch (iInputMode)
		{
		case EAknEditorTextInputMode:
		case EAknEditorKatakanaInputMode:
		case EAknEditorFullWidthTextInputMode:
		case EAknEditorFullWidthNumericInputMode:
		case EAknEditorFullWidthKatakanaInputMode:
		case EAknEditorHiraganaKanjiInputMode:
		case EAknEditorHiraganaInputMode:
		case EAknEditorHalfWidthTextInputMode:
		  if (kana_mode) {
		    response = OfferKeyEventInKanaModeL(aKeyEvent);
		  } else {
		    response = OfferKeyEventInAlphaModeL(aKeyEvent);
		  }
		  break;

		case EAknEditorNumericInputMode:
			response = OfferKeyEventInNumericModeL(aKeyEvent, aEventCode);;
			break;
		
		case EAknEditorSecretAlphaInputMode:
			response = OfferKeyEventInSecretAlphaModeL(aKeyEvent, aEventCode);
			break;
		
		default:
			break;
		}
	return response;
	}

/**
 * Handles keypresses when the fep is in numeric mode
 */
TKeyResponse CSKKmControl::OfferKeyEventInNumericModeL(const TKeyEvent& aKeyEvent, TEventCode /*aEventCode*/)
	{
	TKeyResponse response = EKeyWasNotConsumed;

	if (aKeyEvent.iCode == '*')
		{
		// Just place a key event ('+') to the application underneath the FEP 
		SimulateKeyEventL('+');
		response = EKeyWasConsumed;
		}

	return response;
	}

void CSKKmControl::enter_kana_mode(void)
{
  set_local_capabilities();

#ifdef DEBUG
  print_capabilities();
#endif

  if (capabilities & (TCoeInputCapabilities::EJapaneseHiragana | TCoeInputCapabilities::EAllText)) {
    kana_mode = 1;
    iKanaEngine->set_hiragana();
  } else if (capabilities & TCoeInputCapabilities::EJapaneseKatakanaFullWidth) {
    kana_mode = 1;
    iKanaEngine->set_katakana();
  } else if (capabilities & TCoeInputCapabilities::EJapaneseKatakanaHalfWidth) {
    kana_mode = 1;
    iKanaEngine->set_hankaku();
  } else {
    kana_mode = 0;
    iIndicator->SetState(CFepIndicator::EStateAlpha);
  }
}

/**
 * Handles keypresses when the fep is in kana mode
 */
TKeyResponse CSKKmControl::OfferKeyEventInKanaModeL(const TKeyEvent& aKeyEvent)
{
  TKeyResponse response = EKeyWasNotConsumed;
  TUint keyCode = aKeyEvent.iCode;

  switch (keyCode) {
  case EKeyEnter:
    if (iInsideInlineEditingTransaction) {
      TUint len, commit;
      TPtrC ptr = iKanaEngine->Kakutei(&len, &commit, KAKUTEI_ENTER);
      UpdateInlineEditL(ptr, len);
      if (commit) {
	CommitInlineEditL(*(iInputCapabilities.FepAwareTextEditor()), ETrue);
      }
      abbrev_mode = 0;
      iKanaEngine->set_hiragana();
      response = EKeyWasConsumed;
    } else
      response = EKeyWasNotConsumed;
    break;

  case EKeySpace:
    if (iInsideInlineEditingTransaction) {
      TUint len, commit;
      TPtrC ptr = iKanaEngine->Henkan(&len, &commit);
      UpdateInlineEditL(ptr, len);
      if (commit) {
	CommitInlineEditL(*(iInputCapabilities.FepAwareTextEditor()), ETrue);
      }
      abbrev_mode = 0;
      response = EKeyWasConsumed;
    } else {
      response = EKeyWasNotConsumed;
    }
    break;

  case EKeyBackspace:
    if (iInsideInlineEditingTransaction) {
      TUint len, commit, consume;
      TPtrC ptr = iKanaEngine->backspace(&len, &commit, &consume);
      UpdateInlineEditL(ptr, len);
      if (commit) {
	CommitInlineEditL(*(iInputCapabilities.FepAwareTextEditor()), ETrue);
      }
      if (consume) {
	response = EKeyWasConsumed;
      } else {
	response = EKeyWasNotConsumed;
      }
    } else {
      response = EKeyWasNotConsumed;
    }
    break;

  case EKeyLeftArrow:
  case EKeyRightArrow:
  case EKeyUpArrow:
  case EKeyDownArrow:
    if (iInsideInlineEditingTransaction)
      {
	// If we're in inline editing transaction, just commit it and return
	CommitInlineEditL(*(iInputCapabilities.FepAwareTextEditor()), ETrue);
	abbrev_mode = 0;
	response = EKeyWasConsumed;
      }
    else
      response = EKeyWasNotConsumed;
    break;

  case 0:
    // hack for Freedom Mini
    break;

  default:
    if (keyCode == KEY_CHANGE_ALPHA ||
	keyCode == KEY_CHANGE_ALPHA2 ||
	keyCode == KEY_END_KANA) {
      if (iInsideInlineEditingTransaction) {
	CommitInlineEditL(*(iInputCapabilities.FepAwareTextEditor()), ETrue);
      }
      if (!kana_mode) {
	enter_kana_mode();
      } else {
	kana_mode = 0;
	iIndicator->SetState(CFepIndicator::EStateAlpha);
      }
      abbrev_mode = 0;
      response=EKeyWasConsumed;
    } else if (keyCode == KEY_SAVE_DIC || keyCode == KEY_SAVE_DIC2) {
      iKanaEngine->save_dic();
      response=EKeyWasConsumed;
    } else if (keyCode == ',' || keyCode == '.' || keyCode == '-' ||
	       keyCode == '[' || keyCode == ']' ||
	       keyCode >= 'A' && keyCode <= 'Z' ||
	       keyCode >= 'a' && keyCode <= 'z' ||
	       keyCode == '/') {
      // insert a letter
      HandleKanaModeKeyCodeL(keyCode);
      response=EKeyWasConsumed;
    } else {
      if (iInsideInlineEditingTransaction) {
	TBuf<1> buffer = KNullDesC();
	buffer.Append(keyCode);
	UpdateInlineEditL(buffer, 1);
#ifdef DEBUG
	TBuf<128> buf;
	buf.Format(_L("keycode=%d\n"), keyCode);
	RDebug::Print(buf);
	/*
	TBuf<64> buffer = KNullDesC();
	buf.Format(_L("[%d]"), keyCode);
	UpdateInlineEditL(buf, buf.Length());
	*/
#endif
	CommitInlineEditL(*(iInputCapabilities.FepAwareTextEditor()), ETrue);
	abbrev_mode = 0;
	response = EKeyWasConsumed; 
      } else {
	response = EKeyWasNotConsumed; 
      }
    }
    break;
  }

  return response;
}

/**
 * Handles keypresses when the fep is in secret alpha mode
 */
TKeyResponse CSKKmControl::OfferKeyEventInSecretAlphaModeL(const TKeyEvent& /*aKeyEvent*/, TEventCode /*aEventCode*/)
	{
	// No implementation yet, so just return key not consumed
	return EKeyWasNotConsumed;
	}

/**
 * Handles keypresses when the fep is in alpha mode
 */
TKeyResponse CSKKmControl::OfferKeyEventInAlphaModeL(const TKeyEvent& aKeyEvent)
{
  TKeyResponse response = EKeyWasNotConsumed;
  TUint keyCode = aKeyEvent.iCode;

  switch (keyCode) {
  case KEY_CHANGE_ALPHA:
  case KEY_CHANGE_ALPHA2:
    kana_mode = 1 - kana_mode;
    if (kana_mode) {
      enter_kana_mode();
    } else {
      iIndicator->SetState(CFepIndicator::EStateAlpha);
    }
    response = EKeyWasConsumed;
    break;
  }

  return response;
}

/**
 *
 */
void CSKKmControl::HandleKanaModeKeyCodeL(TUint aKeyCode)
{
  int start_henkan = 0;

  if (!iInsideInlineEditingTransaction) {
    StartInlineEditL(*(iInputCapabilities.FepAwareTextEditor()), KNullDesC, 0, ETrue); // Show cursor
  }
  if (iInsideInlineEditingTransaction) {
    // Update a new character and get a pointer to the engine buffer
    TUint len, commit = 0;
    TUint key = aKeyCode;
    if (iKanaEngine->kanji_mode == 2 && key == KEY_REMOVE_ENTRY) {
      TPtrC ptr = iKanaEngine->remove_entry(&len);
      UpdateInlineEditL(ptr, len); // multi character
      commit = 0;
    } else {
      if (key >= 'A' && key <= 'Z' && !abbrev_mode && key != KEY_TOGGLE_HANKAKU) {
	key = key - 'A' + 'a';
	if (iKanaEngine->kanji_mode == 1) {
	  start_henkan = 1;
	} else {
	  iKanaEngine->Kakutei(&len, &commit, KAKUTEI_MAJOR);
	  if (commit) {
	    iKanaEngine->SetKakutei();
	  }
	  iKanaEngine->set_hiragana();
	  abbrev_mode = 0;
	  ClearBuffer();
	  iKanaEngine->KanjiIn();
	}
      }
      if (iKanaEngine->kanji_mode == 2 && aKeyCode == KEY_HENKAN_BACKWARD) {
	TPtrC ptr = iKanaEngine->henkan_backward(&len);
	UpdateInlineEditL(ptr, len); // multi character
	commit = 0;
      } else if (aKeyCode == KEY_START_ABBREV) {
	// abbrev mode
	if (!abbrev_mode) {
	  //	  if (iKanaEngine->kanji_mode) {
	    iKanaEngine->Kakutei(&len, &commit, KAKUTEI_MAJOR);
	    if (commit) {
	      iKanaEngine->SetKakutei();
	    }
	    //	  }
	  iKanaEngine->set_hiragana();
	  abbrev_mode = 1;
	  iKanaEngine->KanjiIn();
	  iKanaEngine->PrepareRetbuf();
	  TPtrC ptr = iKanaEngine->Retbuf();
	  len = 1;
	  UpdateInlineEditL(ptr, len);
	} else {
	  TPtrC ptr = iKanaEngine->AppendAbbrev(key, &len);
	  UpdateInlineEditL(ptr, len);
	}
	commit = 0;
      } else {
	if (start_henkan) {
	  iKanaEngine->AppendCharacter(key, &len, &commit, SET_POST);
	  TPtrC ptr = iKanaEngine->Henkan(&len, &commit);
	  UpdateInlineEditL(ptr, len); // multi character
	} else if (abbrev_mode) {
	  TPtrC ptr = iKanaEngine->AppendAbbrev(key, &len);
	  UpdateInlineEditL(ptr, len); // multi character
	  commit = 0;
	} else if (aKeyCode == KEY_TOGGLE_KATAKANA) {
	  set_local_capabilities();
#ifdef DEBUG
	  print_capabilities();
#endif
	  iKanaEngine->toggle_katakana(capabilities);
	} else if (aKeyCode == KEY_TOGGLE_HANKAKU) {
	  set_local_capabilities();
#ifdef DEBUG
	  print_capabilities();
#endif
	  iKanaEngine->toggle_hankaku(capabilities);
	} else {
	  TPtrC ptr = iKanaEngine->AppendCharacter(key, &len, &commit, NO_SET);
	  UpdateInlineEditL(ptr, len); // multi character
	}
      }
    }
    if (commit) {
      CommitInlineEditL(*(iInputCapabilities.FepAwareTextEditor()), ETrue);
    }
  }
}

/**
 * This method will place one key event defined by aKeyCode onto the control
 * stack to be handled as normal.
 */
void CSKKmControl::SimulateKeyEventL(TUint aKeyCode)
	{
	CArrayFix<TUint>* simCharArray=new(ELeave) CArrayFixFlat<TUint>(1);
	CleanupStack::PushL(simCharArray);
	simCharArray->AppendL(aKeyCode);
	iFep.SimulateKeyEventsL(simCharArray->Array());
	CleanupStack::PopAndDestroy(); // simCharArray
	}


void CSKKmControl::HandleChangeInFocus()
{
  TRAPD(err, HandleChangeInFocusL());
  if (err != KErrNone) {
    ClearBuffer();
    ClearKakutei();
  }
}

void CSKKmControl::HandleChangeInFocusL()
{
  // Commit the uncommitted inline editing, if any
  if ((iInputMode == EAknEditorTextInputMode || iInputMode == EAknEditorNumericInputMode ||
       iInputMode == EAknEditorSecretAlphaInputMode) && iInsideInlineEditingTransaction)
    {
      CommitInlineEditL(*(iInputCapabilities.FepAwareTextEditor()), EFalse);
    }

  // Get the newly focused control capability	
  iInputCapabilities = STATIC_CAST(const CCoeAppUi*, iCoeEnv->AppUi())->InputCapabilities();

  // check & process the current capability underneath
  if (iInputCapabilities.FepAwareTextEditor())
    {
      // Not all text editors are "fep aware". Fep somehow needs help from the editor window 
      // to determine the current editor capabilities.
      CAknEdwinState* editorState = EditorState();
      TUint editorMode = editorState->CurrentInputMode();
      if (editorMode == EAknEditorNullInputMode)
	editorMode = editorState->DefaultInputMode();

      SwitchInputModeL(editorMode);
#ifdef DEBUG
      RDebug::Print(_L("fep aware:"));
      print_capabilities();
#endif
    }
  else // no FepAwareTextEditor
    { // Check for non Edwin derived editors (eg mfne, tel no editor etc)
#ifdef DEBUG
      RDebug::Print(_L("no fep aware:"));
      print_capabilities();
#endif
      if (iInputCapabilities.SupportsWesternNumericIntegerPositive() || iInputCapabilities.SupportsWesternNumericIntegerNegative())
	{
	  SwitchInputModeL(EAknEditorNumericInputMode);
	}
      else if (iInputCapabilities.SupportsSecretText())
	{
	  SwitchInputModeL(EAknEditorSecretAlphaInputMode);
	}
      else if (iInputCapabilities.SupportsWesternAlphabetic())
	{
	  SwitchInputModeL(EAknEditorTextInputMode);
	}
      else 
	{
	  SwitchInputModeL(EAknEditorNullInputMode);
	}
    }

  if (iInputMode & (EAknEditorTextInputMode |
		    EAknEditorKatakanaInputMode |
		    EAknEditorFullWidthTextInputMode |
		    EAknEditorFullWidthNumericInputMode |
		    EAknEditorFullWidthKatakanaInputMode |
		    EAknEditorHiraganaKanjiInputMode |
		    EAknEditorHiraganaInputMode |
		    EAknEditorHalfWidthTextInputMode)) {
    enter_kana_mode();
  }
}

CAknEdwinState* CSKKmControl::EditorState() const
	{
	return STATIC_CAST(CAknEdwinState*, iInputCapabilities.FepAwareTextEditor()->Extension1()->State(KNullUid));
	}

/*
 * This method switch and set the input mode of the Avkon editor state
 */
void CSKKmControl::SwitchInputModeL(TUint aInputMode)
{
  iInputMode = aInputMode;
  switch(aInputMode)
    {
    case EAknEditorTextInputMode:			// Fall through
    case EAknEditorSecretAlphaInputMode:	
      if (kana_mode) {
	iIndicator->SetState(CFepIndicator::EStateKana);
      } else {
	iIndicator->SetState(CFepIndicator::EStateAlpha);
      }
      break;

    case EAknEditorNumericInputMode:
      //	iIndicator->SetState(CFepIndicator::EStateNumeric);
      break;

    case EAknEditorNullInputMode:
      iIndicator->SetState(CFepIndicator::EStateNone);
      // Newly focused control is null input mode, reset flag
      iInsideInlineEditingTransaction = EFalse;			// Update indicator on status pane
      break;

    default:
      break;
    }

  ClearBuffer();
  ClearKakutei();

  // ensure editor is aware of new fep mode
  if (iInputCapabilities.FepAwareTextEditor())
    EditorState()->SetCurrentInputMode(aInputMode);
}

void CSKKmControl::ClearBuffer(void)
{
  iKanaEngine->ClearBuffer();
}

void CSKKmControl::ClearKakutei(void)
{
  iKanaEngine->ClearKakutei();
}

void CSKKmControl::set_local_capabilities(void)
{
  capabilities = iInputCapabilities.Capabilities();
  switch (iInputMode) {
  case EAknEditorKatakanaInputMode:
    capabilities = TCoeInputCapabilities::EJapaneseKatakanaHalfWidth;
    break;
  }
}

#ifdef DEBUG
void CSKKmControl::print_capabilities(void)
{
  TBuf<128> buf;
  buf.Format(_L("InputMode=%x caps=%x\n"),
	     iInputMode,capabilities);
  RDebug::Print(buf);
}
#endif
