
#include <eikenv.h>         //CEikEnv
#include <eikspane.h>       //CEikStatusPane
#include <avkon.hrh>
#include <aknnavi.h>        //CAknNavigationControlContainer
#include <aknnavide.h>      //CAknNavigationDecorator
#include <aknindicatorcontainer.h>  //CAknIndicatorContainer
#include <aknutils.h>
#include "FepIndicator.h"
#include "utf.h"

_LIT8(KNaviPaneTextKana, "[\xe3\x81\x8b\xe3\x81\xaa]"); // 
_LIT8(KNaviPaneTextKatakana, "[\xe3\x82\xab\xe3\x83\x8a]"); // Ji
_LIT8(KNaviPaneTextHankaku, "[\xef\xbd\xb6\xef\xbe\x85]"); // 
_LIT(KNaviPaneTextAlpha, "[alph]");

/**
 *
 */
CFepIndicator* CFepIndicator::NewL()
    {
    CFepIndicator* self = new(ELeave)CFepIndicator();
    return self;
    }


/**
 *
 */
CFepIndicator::CFepIndicator()
    :iPreviousState(EStateNone),
	iEikEnv(*CEikonEnv::Static())
    {
    }

/**
 *
 */
CFepIndicator::~CFepIndicator()
    {
    RemoveContainer();
    }


/**
 *
 */
void CFepIndicator::SetState(TIndicatorState aState)
{
  TBuf<64> tmp2 = KNullDesC();
  CAknNavigationControlContainer* naviPane = 0;
  TRAPD(err, naviPane = NaviPaneL());

  // If the navi pane has been changed, clean up and start again.
  if (iNaviPane != naviPane)
    {
      RemoveContainer();
      iNaviPane = naviPane;
      iNaviPaneChanged = ETrue;
    }

  if ( aState != EStateLabel && (iPreviousState == aState) && !iNaviPaneChanged ) {
    // The required state already exists and navi pane has not been changed
    // by launching embedded application. Leave the contents alone.
    return;
  }

  if (aState == EStateNone)
    {
      if ( iNaviDecorator && naviPane )
	{
	  naviPane->Pop(iNaviDecorator);
	}
      iPreviousState = aState;
      iNaviPaneChanged = EFalse;
      return;
    }

  if ( err == KErrNone ) {
    iNaviPaneChanged = EFalse;
    TRAP(err,AddContainerL(naviPane));
    // Don't propagate the error since the only side-effect is that
    // the edit indicators won't show up

    switch (aState) {
    case EStateAlpha:
      if (iNaviDecorator) {
	delete iNaviDecorator;
	iNaviDecorator = NULL;
      }

      iNaviDecorator = iNaviPane->CreateMessageLabelL(KNaviPaneTextAlpha);
      // Push the state to the navi-pane
      if (naviPane) {
	TRAPD(ignore, naviPane->PushL(*iNaviDecorator));
      }
      break;

    case EStateKana:
      if (iNaviDecorator) {
	delete iNaviDecorator;
	iNaviDecorator = NULL;
      }

      CnvUtfConverter::ConvertToUnicodeFromUtf8(tmp2, KNaviPaneTextKana);
      iNaviDecorator = iNaviPane->CreateMessageLabelL(tmp2);
      // Push the state to the navi-pane
      if (naviPane) {
	TRAPD(ignore, naviPane->PushL(*iNaviDecorator));
      }
      break;

    case EStateKatakana:
      if (iNaviDecorator) {
	delete iNaviDecorator;
	iNaviDecorator = NULL;
      }

      CnvUtfConverter::ConvertToUnicodeFromUtf8(tmp2, KNaviPaneTextKatakana);
      iNaviDecorator = iNaviPane->CreateMessageLabelL(tmp2);
      // Push the state to the navi-pane
      if (naviPane) {
	TRAPD(ignore, naviPane->PushL(*iNaviDecorator));
      }
      break;

    case EStateHankaku:
      if (iNaviDecorator) {
	delete iNaviDecorator;
	iNaviDecorator = NULL;
      }

      CnvUtfConverter::ConvertToUnicodeFromUtf8(tmp2, KNaviPaneTextHankaku);
      iNaviDecorator = iNaviPane->CreateMessageLabelL(tmp2);
      // Push the state to the navi-pane
      if (naviPane) {
	TRAPD(ignore, naviPane->PushL(*iNaviDecorator));
      }
      break;

    case EStateLabel:
      if (iNaviDecorator) {
	delete iNaviDecorator;
	iNaviDecorator = NULL;
      }

      iNaviDecorator = iNaviPane->CreateMessageLabelL(state_label);
      // Push the state to the navi-pane
      if (naviPane) {
	TRAPD(ignore, naviPane->PushL(*iNaviDecorator));
      }
      break;

    }
  }
  iPreviousState = aState;
}


/**
 *  Create a new indicator container and push it to the navi-pane
 */
void CFepIndicator::AddContainerL(CAknNavigationControlContainer* aNaviPane)
    {
    // Create a new indicator container and push it to the navi-pane
    if (!iIndicatorContainer && aNaviPane)
        {
        iNaviDecorator = aNaviPane->CreateEditorIndicatorContainerL();
        iIndicatorContainer = STATIC_CAST(CAknIndicatorContainer*, iNaviDecorator->DecoratedControl());
        }
    }

/**
 * Pop the indicator container from the navi-pane
 */
void CFepIndicator::RemoveContainer()
    {
    // Pop the indicator container from the navi-pane
    if (iIndicatorContainer)
        {
        delete iNaviDecorator;
        iNaviDecorator = NULL;
        iIndicatorContainer = NULL;
        }
    }

/**
 *  get a handle to the Navi Pane
 */
CAknNavigationControlContainer* CFepIndicator::NaviPaneL()
    {
    CEikStatusPane *sp = iEikEnv.AppUiFactory()->StatusPane();
    User::LeaveIfNull(sp);
    // Fetch pointer to the default navi pane control
    return (CAknNavigationControlContainer *)sp->ControlL(TUid::Uid(EEikStatusPaneUidNavi));
    }

void CFepIndicator::set_label(TDesC& label)
{
  state_label.FillZ();
  state_label.Copy(label);
  SetState(EStateLabel);
}

// End of file
