#include <utf.h>
#include <f32file.h>
#include <coemain.h>
#include <coeinput.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/unistd.h>

#include "kanaengine.h"
#include "FepIndicator.h"

const char dicdrives[] = "cdefghijklmnopqrstuvwxyzab";

#define DICDIR "\\system\\dic\\"

#define DICFILE DICDIR "skk.dic"
#define NEWFILE DICDIR "skk.new"
#define OLDFILE DICDIR "skk.old"

#define MARK_REG		_L8("\xe3\x80\x8b") // "t"
#define MARK_IN_CONV_REG	_L8("\xe2\x96\xa0") // ""
#define MARK_IN_CONV		_L8("\xe2\x96\xbc") // ""
#define MARK_START_CONV_REG	_L8("\xe2\x96\xa1") // ""
#define MARK_START_CONV		_L8("\xe2\x96\xbd") // ""

#define HIRAGANA_TT _L8("\xe3\x81\xa3") // 
#define HIRAGANA_NN _L8("\xe3\x82\x93") // 
#define KATAKANA_TT _L8("\xe3\x83\x83") // b
#define KATAKANA_NN _L8("\xe3\x83\xb3") // 
#define HANKAKU_TT  _L8("\xef\xbd\xaf") // 
#define HANKAKU_NN  _L8("\xef\xbe\x9d") // 

#define REG_LABEL _L8("\xe7\x99\xbb:") // o:

struct struct_kana_map {
  char *key, *val;
};

const struct struct_kana_map  kana_map[] = {
#include "map_hiragana.h"
};


const struct struct_kana_map  katakana_map[] = {
#include "map_katakana.h"
};

const struct struct_kana_map  hankaku_map[] = {
#include "map_hankaku.h"
};


DicPtr::DicPtr(){
  next = 0;
}

void DicPtr::NewContents(RBuf &b){
  contents.CreateL(b);
}

void DicPtr::NewContents(TDesC16 &b){
  contents.CreateL(b);
}

void DicPtr::UpdateContents(RBuf &b){
  contents.ReAllocL(b.Length()+2);
  contents.Copy(b);
}

void DicPtr::SetNext(DicPtr *n){
  next = n;
}

//////////////////////////////////////////////////////////////////////
//
// Class CSKKmKanaEngine
//
//////////////////////////////////////////////////////////////////////
CSKKmKanaEngine* CSKKmKanaEngine::NewL(CSKKmControl& aControl, CFepIndicator* aIndicator)
	{
	CSKKmKanaEngine* self = new(ELeave)CSKKmKanaEngine(aControl, aIndicator);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();	// self
	return self;
	}

CSKKmKanaEngine::CSKKmKanaEngine(CSKKmControl& aControl, CFepIndicator* aIndicator)
  : kanji_mode(0),
    iFepControl(aControl),
    iIndicator(aIndicator),
    iBuffer(KNullDesC),
    henkan_candidate(0),
    yomi(KNullDesC),
    kanji(KNullDesC),
    kakutei(KNullDesC),
    retbuf(KNullDesC),
    regbuf(KNullDesC),
    first((DicPtr *)0),
    cur_reg_entry(0),
    post_char(0),
    katakana(0),
    hankaku(0)
{
}

void CSKKmKanaEngine::ConstructL()
{
  for(int i=0; i<BUFSIZE; i++) {
    prev_strs[i] = 0;
  }
  n_conv = sizeof(kana_map) / sizeof(struct struct_kana_map);

  //  load_dic();

  CnvUtfConverter::ConvertToUnicodeFromUtf8(mark_registering, MARK_REG);

  CnvUtfConverter::ConvertToUnicodeFromUtf8(mark_in_conversion_reg, MARK_IN_CONV_REG);

  CnvUtfConverter::ConvertToUnicodeFromUtf8(mark_in_conversion, MARK_IN_CONV);

  CnvUtfConverter::ConvertToUnicodeFromUtf8(mark_start_conversion_reg, MARK_START_CONV_REG);

  CnvUtfConverter::ConvertToUnicodeFromUtf8(mark_start_conversion, MARK_START_CONV);
}

CSKKmKanaEngine::~CSKKmKanaEngine()
{
  DicPtr *q = first;
  DicPtr *r;

  if (first) {
    while(q->next){
      r = q;
      q = q->next;
      delete r;
    }
  }
}

void CSKKmKanaEngine::load_dic()
{
  FILE *fp;
  char buf[BUFSIZE];
  char *d;

  dicdrive = 0;
  for(d=(char *)dicdrives; *d && !dicdrive ; d++){
    buf[0] = *d;
    buf[1] = ':';
    buf[2] = 0;
    strncat(buf, DICFILE, BUFSIZE);

    first = new DicPtr;
    DicPtr *q = first;

    if ((fp = fopen(buf, "r"))) {
      dicdrive = *d;

      while(!feof(fp)){
	char *e;
	if ((e = fgets(buf, BUFSIZE,fp))) {
	  int l = strlen(buf);
	  if (buf[l-1] < ' ') {
	    buf[l-1] = 0;
	  }
	  if (buf[l-2] < ' ') {
	    buf[l-2] = 0;
	  }
	  utfbuf.Copy((TUint8 *)buf);
	  CnvUtfConverter::ConvertToUnicodeFromUtf8(ucsbuf, utfbuf);
	  q->NewContents(ucsbuf);
	  DicPtr *r = new DicPtr;
	  q->SetNext(r);
	  q = r;
	}
      }
      fclose(fp);
    }
  }
}


void CSKKmKanaEngine::save_dic()
{
  FILE *fp;
  char *newfile, *oldfile, *dicfile;

  if (first) {
    newfile = (char *) malloc(BUFSIZE);
    newfile[0] = dicdrive;
    newfile[1] = ':';
    newfile[2] = 0;
    strncat(newfile, NEWFILE, BUFSIZE);

    DicPtr *q = first;

    fp = fopen(newfile, "w");

    while(q->next){
      ucsbuf.Copy(q->contents);
      ucsbuf.ZeroTerminate();
      CnvUtfConverter::ConvertFromUnicodeToUtf8(utfbuf, ucsbuf);
      utfbuf.ZeroTerminate();
      fprintf(fp, "%s\n", utfbuf.Ptr());
      q = q->next;
    }

    fclose(fp);

    oldfile = (char *) malloc(BUFSIZE);
    oldfile[0] = dicdrive;
    oldfile[1] = ':';
    oldfile[2] = 0;
    strncat(oldfile, OLDFILE, BUFSIZE);

    dicfile = (char *) malloc(BUFSIZE);
    dicfile[0] = dicdrive;
    dicfile[1] = ':';
    dicfile[2] = 0;
    strncat(dicfile, DICFILE, BUFSIZE);
    rename(dicfile, oldfile);
    rename(newfile, dicfile);
    free(newfile);
    free(oldfile);
    free(dicfile);
  }
}

TPtrC CSKKmKanaEngine::AppendAbbrev(TUint aKeyCode, TUint *len)
{
  iBuffer.Append(aKeyCode);
  PrepareRetbuf();
  *len = retbuf.Length();
  return retbuf.Ptr();
}

TPtrC CSKKmKanaEngine::AppendCharacter(TUint aKeyCode, TUint *len, TUint *commit, TUint set_post)
{
  int i,l;
  char *found_str = NULL;
  int can_continue = 0;
  TBuf<64> buffer = KNullDesC();
  TBuf<64> tmp2 = KNullDesC();
  TBuf8<64> tmp1 = KNullDesC8();

  const struct struct_kana_map *map;

  if (set_post) {
    post_char = aKeyCode;
  }

  *commit = 0;

  l = strlen(prev_strs);

  if (!set_post && post_char) {
    iBuffer.Append(post_char);
    prev_strs[l] = (char) post_char;
    l++;
    post_char = 0;
  }

  prev_strs[l] = (char) aKeyCode;
  prev_strs[l+1] = 0;

  if (hankaku) {
    map = hankaku_map;
  } else if (katakana) {
    map = katakana_map;
  } else {
    map = kana_map;
  }

  for(i=0; i<n_conv; i++){
    if (!strcmp(prev_strs, map[i].key)) {
      found_str = map[i].val;
    } else if (!l || !strncmp(prev_strs, map[i].key, l)) {
      can_continue = 1;
    }
  }

  if (found_str) {
    for(i=0; i<l; i++){
      // Delete the previous letter in iBuffer
      DeleteCharacter();
    }

    tmp1.Copy((TUint8 *)found_str);
    CnvUtfConverter::ConvertToUnicodeFromUtf8(tmp2, tmp1);
    buffer.Append(tmp2);

    prev_strs[0] = 0;
    if (!kanji_mode && !cur_reg_entry) {
      *commit = COMMIT_COMMIT;
    }
  } else {
    if (l && prev_strs[l-1] == (char) aKeyCode) {
      DeleteCharacter();
      if (hankaku) {
	CnvUtfConverter::ConvertToUnicodeFromUtf8(tmp2, HANKAKU_TT);
      } else if (katakana) {
	CnvUtfConverter::ConvertToUnicodeFromUtf8(tmp2, KATAKANA_TT);
      } else {
	CnvUtfConverter::ConvertToUnicodeFromUtf8(tmp2, HIRAGANA_TT);
      }
      buffer.Copy(tmp2);
      buffer.Append(aKeyCode);
      prev_strs[l] = 0;
    } else if (l && prev_strs[l-1] == 'n' && aKeyCode != 'y') {
      DeleteCharacter();
      if (hankaku) {
	CnvUtfConverter::ConvertToUnicodeFromUtf8(tmp2, HANKAKU_NN);
      } else if (katakana) {
	CnvUtfConverter::ConvertToUnicodeFromUtf8(tmp2, KATAKANA_NN);
      } else {
	CnvUtfConverter::ConvertToUnicodeFromUtf8(tmp2, HIRAGANA_NN);
      }
      buffer.Copy(tmp2);
      buffer.Append(aKeyCode);
      prev_strs[l-1] = (char) aKeyCode;
      prev_strs[l] = 0;
    } else {
      buffer.Append(aKeyCode);
      if (can_continue) {
      } else {
	if (!kanji_mode && !cur_reg_entry) {
	  *commit = COMMIT_COMMIT;
	}
        prev_strs[0] = 0;
      }
    }
  }

  iBuffer += buffer;

  PrepareRetbuf();
  *len = retbuf.Length();
  return retbuf.Ptr();
}

DicPtr *CSKKmKanaEngine::search_entry()
{
  DicPtr *q = first;
  DicPtr *prev_dicptr = first;
  RBuf a;

  a.CreateL(BUFSIZE);
  a.Copy(yomi);
  a.Append(' ');

  while(q->next){
    int m = q->contents.Find(a);
    if (!m) {
      break;
    }
    prev_dicptr = q;
    q = q->next;
  }

  if (q->next) {
    // found entry
    // sort entry
    if (prev_dicptr != first) {
      prev_dicptr->SetNext(q->next);
      q->SetNext(first);
      first = q;
    }
  } else {
    // not found entry
    q = 0;
  }

  return q;
}

int CSKKmKanaEngine::candidate(DicPtr *q, int n){
  int i = 0;
  int err = 0;
  RBuf t;
  int sl1, sl2;

  t.CreateL(q->contents);
  c_begin = 0;

  while(i<n) {
    sl1 = t.Locate('/');
    c_begin += (sl1+1);
    if (sl1<0) {
      err = -1;
    } else {
      RBuf u;
      u.CreateL(t.Right(t.Length() - sl1 - 1));
      t.FillZ();
      t.Copy(u);
      i++;
    }
  }

  if (!t.Length()) {
    err = -1;
  }

  if (err) {
    //    printf("candidate [%d] not found.\n",n);
  } else {
    sl2 = t.Locate('/');

    kanji.Copy(t.Left(sl2));
    c_end = c_begin + kanji.Length();
  }

  return err;
}

void CSKKmKanaEngine::set_frequent(DicPtr *q, int n)
{
  RBuf a,b,c;

  candidate(q,n);
  int sp = q->contents.Locate('/');

  a.CreateL(BUFSIZE);
  a.Copy(q->contents.Left(sp+1));
  b.CreateL(q->contents.Mid(sp+1, c_begin - sp - 1));
  c.CreateL(q->contents.Right(q->contents.Length() - c_end - 1));

  a.Append(kanji);
  a.Append('/');
  a.Append(b);
  a.Append(c);

  q->contents.Copy(a);
}

TUint CSKKmKanaEngine::SearchKanji(int n)
{
  // search kanji using (yomi, n) -> set kanji
  // return 0 if n exceeds
  // if yomi is not found, set yomi to kanji

  if (!n) {
    found_entry = search_entry();
  }

  kanji.FillZ();
  kanji.SetLength(0);

  if (found_entry) {
    int err = candidate(found_entry, n+1);
    if (err) {
      // register new entry
      n = -2;
      // circulate entry
      //      n = 0;
      //      candidate(found_entry, n+1);
    } else {
      //      kanji.Append(found_entry->contents);
    }
  } else {
    n = -1;
  }

#ifdef DEBUG
  if (n<0) {
    TBuf<64> tmp2 = KNullDesC();
    TBuf8<64> tmp1 = KNullDesC8();
    tmp1.Copy((TUint8 *)"(NO CANDIDATE)");
    CnvUtfConverter::ConvertToUnicodeFromUtf8(tmp2, tmp1);
    kanji.Num(n);
    kanji.Append(':');
    kanji.Append(tmp2);

    //kanji.Append(yomi);
  }
#endif
  return n;
}

void CSKKmKanaEngine::create_new_kanji_entry(RegEntry* e)
{
  DicPtr *q = new DicPtr;
  RBuf a;

  a.CreateL(BUFSIZE);
  a.Copy(e->yomi);
  a.Append(_L(" /"));
  a.Append(e->kanji);
  a.Append('/');

  q->NewContents(a);
  q->next = first;

  first = q;
}

void CSKKmKanaEngine::append_kanji(RegEntry* e)
{
  DicPtr *q = e->entry;
  RBuf a;

  int sp = q->contents.Locate('/');

  a.CreateL(BUFSIZE);
  a.Copy(q->contents.Left(sp+1));
  a.Append(e->kanji);

  int l = q->contents.Length() - sp;
  a.Append(q->contents.Right(l));

  q->UpdateContents(a);
}

void CSKKmKanaEngine::start_reg_kanji()
{
  RegEntry *q;

  q = new RegEntry;
  q->yomi.CreateL(yomi);
  q->entry = found_entry;
  q->kanji.CreateL(BUFSIZE);
  q->regbuf.CreateL(regbuf);
  q->prev = cur_reg_entry;
  q->post_char = post_char;

  post_char = 0;

  CnvUtfConverter::ConvertToUnicodeFromUtf8(ucsbuf, REG_LABEL);
  ucsbuf.Append(yomi);
  q->label.CreateL(ucsbuf);

  iIndicator->set_label(q->label);

  cur_reg_entry = q;
}

void CSKKmKanaEngine::commit_reg_kanji()
{
  RegEntry *q = cur_reg_entry;

  if (q->entry) {
    append_kanji(q);
  } else {
    create_new_kanji_entry(q);
  }

  regbuf.FillZ();
  regbuf.Copy(q->regbuf);

  post_char = q->post_char;

  cur_reg_entry = q->prev;
  if (cur_reg_entry) {
    iIndicator->set_label(cur_reg_entry->label);
  } else {
    set_hiragana();
  }

  delete q;

}

void CSKKmKanaEngine::cancel_reg_kanji()
{
  RegEntry *q = cur_reg_entry;

  regbuf.FillZ();
  regbuf.Copy(q->regbuf);

  //  post_char = q->post_char;
  post_char = 0;

  cur_reg_entry = q->prev;
  if (cur_reg_entry) {
    iIndicator->set_label(cur_reg_entry->label);
  } else {
    set_hiragana();
  }

  delete q;
}

TPtrC CSKKmKanaEngine::remove_entry(TUint *len)
{
  DicPtr *q = found_entry;
  int n = henkan_candidate;
  RBuf a,c;

  candidate(q,n+1);

  a.CreateL(BUFSIZE);
  a.Copy(q->contents.Left(c_begin));

  c.CreateL(q->contents.Right(q->contents.Length() - c_end - 1));
  a.Append(c);

  q->contents.Copy(a);

  if (candidate(q,1)) {
    // remove whole entry
    DicPtr *prev = first;
    for(q=first; q!=found_entry; q=q->next){
      prev = q;
    }
    if (prev == first) {
      first = q->next;
      delete q;
    } else {
      prev->next = q->next;
      delete q;
    }
  }

  ClearBuffer();
  kanji_mode = 0;

  PrepareRetbuf();
  *len = retbuf.Length();
  return retbuf.Ptr();
}

TPtrC CSKKmKanaEngine::Henkan(TUint *len, TUint *commit)
{
  TBuf<64> tmp2;

  if (!first) {
    load_dic();
  }

  if (cur_reg_entry) {
    tmp2.Copy(mark_in_conversion_reg);
  } else {
    tmp2.Copy(mark_in_conversion);
  }

  TUint l = iBuffer.Length() - 1;
  *commit = 0;

  switch (kanji_mode) {
  case 1:
    // start henkan
    kanji_mode = 2;
    yomi.FillZ();
    yomi.SetLength(0);
    yomi.Append(iBuffer.Right(l));
    henkan_candidate = SearchKanji(0);
    ClearBuffer();
    if (henkan_candidate>=0) {
      iBuffer.Append(tmp2);
      iBuffer.Append(kanji);
    } else {
      // register whole entry
      found_entry=0;
      start_reg_kanji();
      ClearRegbuf();
      regbuf.Copy(mark_registering);
      kanji_mode = 0;
      *commit = 0;
    }
    break;

  case 2:
    // in henkan
    henkan_candidate ++;
    henkan_candidate = SearchKanji(henkan_candidate);
    ClearBuffer();
    if (henkan_candidate>=0) {
      iBuffer.Append(tmp2);
      iBuffer.Append(kanji);
    } else {
      // enter register new kanji mode
      start_reg_kanji();
      ClearRegbuf();
      regbuf.Copy(mark_registering);
      kanji_mode = 0;
      *commit = 0;
    }
    break;

  case 0:
  default:
    if (cur_reg_entry) {
      *commit = 0;
    } else {
      // insert space
      iBuffer.Append(' ');
      *commit = COMMIT_COMMIT;
    }
  }

  PrepareRetbuf();
  *len = retbuf.Length();
  return retbuf.Ptr();
}

TPtrC CSKKmKanaEngine::henkan_backward(TUint *len)
{
  TBuf<64> tmp2;

  if (cur_reg_entry) {
    tmp2.Copy(mark_in_conversion_reg);
  } else {
    tmp2.Copy(mark_in_conversion);
  }

  if (henkan_candidate) {
    henkan_candidate --;
  }
  henkan_candidate = SearchKanji(henkan_candidate);
  ClearBuffer();
  if (henkan_candidate>=0) {
    iBuffer.Append(tmp2);
    iBuffer.Append(kanji);
  }

  PrepareRetbuf();
  *len = retbuf.Length();
  return retbuf.Ptr();
}

TPtrC CSKKmKanaEngine::Kakutei(TUint *len, TUint *commit, TUint mode)
{
  *len = iBuffer.Length();
  *commit = COMMIT_COMMIT;

  prev_strs[0] = 0;

  if (kanji_mode) {
    if (kanji_mode == 2) {
      set_frequent(found_entry, henkan_candidate+1);
    }
    kanji_mode = 0;
    if (*len>1) {
      *len -= 1;
      ucsbuf.Copy(iBuffer.Right(*len));
      ClearBuffer();
      if (cur_reg_entry) {
	cur_reg_entry->kanji.Append(ucsbuf);
	regbuf.Append(ucsbuf);
	*commit = 0;
      } else {
	iBuffer.Append(ucsbuf);
      }
    }
  } else if (cur_reg_entry) {
    // kakutei registering entry or first input of major char
    if (*len) {
      // append kana
      cur_reg_entry->kanji.Append(iBuffer);
      regbuf.Append(iBuffer);
    }

    ClearBuffer();

    if (mode == KAKUTEI_MAJOR) {
      *commit = 0;
    } else {
      if (cur_reg_entry->kanji.Length()) {
	ucsbuf2.Copy(cur_reg_entry->kanji);
	commit_reg_kanji();
	if (cur_reg_entry) {
	  cur_reg_entry->kanji.Append(ucsbuf2);
	  regbuf.Append(ucsbuf2);
	  *commit = 0;
	} else {
	  iBuffer.Copy(ucsbuf2);
	  if (post_char) {
	    *commit = 0;
	    TUint code = post_char;
	    post_char = 0;
	    TUint dummy_len, dummy_commit;
	    AppendCharacter(code, &dummy_len, &dummy_commit, 0);
	  }
	}
      } else {
	cancel_reg_kanji();
	if (cur_reg_entry) {
	  *commit = 0;
	}
      }
    }
  }

  PrepareRetbuf();
  *len = retbuf.Length();
  return retbuf.Ptr();
}

TPtrC CSKKmKanaEngine::backspace(TUint *len, TUint *commit, TUint *consume)
{
  int l = iBuffer.Length();
  *commit = 0;
  *consume = 1;

  switch (kanji_mode) {
  case 0: // kana
    if (post_char) {
      post_char = 0;
    }
    if (cur_reg_entry) {
      if (l) {
	DeleteCharacter();
	remove_prev_strs();
      } else {
	if ((l = cur_reg_entry->kanji.Length())) {
	  ucsbuf.Copy(cur_reg_entry->kanji.Left(l-1));
	  cur_reg_entry->kanji.Copy(ucsbuf);

	  if ((l = regbuf.Length())) {
	    ucsbuf.Copy(regbuf.Left(l-1));
	    regbuf.FillZ();
	    regbuf.Copy(ucsbuf);
	  }
	} else {
	  cancel_reg_kanji();
	}
      }
    } else {
      if (l) {
	DeleteCharacter();
	remove_prev_strs();
      } else {
	*commit = COMMIT_COMMIT;
	*consume = 0;
      }
    }
    break;

  case 1: // yomi input
    if (post_char) {
      post_char = 0;
    }
    if (l>=2) {
      DeleteCharacter();
      remove_prev_strs();
    }
    if (l<=2) {
      kanji_mode = 0;
      ClearBuffer();
      remove_prev_strs();
    }
    break;

  case 2: // in henkan
    Kakutei(len, commit, KAKUTEI_ENTER);
    if (post_char) {
      post_char = 0;
    } else if (cur_reg_entry) {
      if ((l = cur_reg_entry->kanji.Length())) {
	ucsbuf.Copy(cur_reg_entry->kanji.Left(l-1));
	cur_reg_entry->kanji.Copy(ucsbuf);
      }
      if ((l = regbuf.Length())) {
	ucsbuf.Copy(regbuf.Left(l-1));
	regbuf.Copy(ucsbuf);
      }
    } else {
      DeleteCharacter();
    }
    break;
  }

  PrepareRetbuf();
  *len = retbuf.Length();
  return retbuf.Ptr();
}

void CSKKmKanaEngine::SetKakutei()
{
  kakutei.FillZ();
  kakutei.Copy(iBuffer);
}

void CSKKmKanaEngine::ClearKakutei()
{
  kakutei.FillZ();
  kakutei.SetLength(0);
}

void CSKKmKanaEngine::KanjiIn()
{
  kanji_mode = 1;

  ClearBuffer();
  if (cur_reg_entry) {
    iBuffer.Append(mark_start_conversion_reg);
  } else {
    iBuffer.Append(mark_start_conversion);
  }
}

TPtrC CSKKmKanaEngine::DeleteCharacter(void)
	{
	TInt len = iBuffer.Length();
          if (len>0) {
            iBuffer = iBuffer.Left (len - 1);
          }
	return iBuffer.Ptr();
	}

void CSKKmKanaEngine::ClearBuffer(void)
{
  iBuffer.FillZ();
  iBuffer.SetLength(0);
  prev_strs[0] = 0;
}

void CSKKmKanaEngine::ClearRegbuf(void)
{
  regbuf.FillZ();
  regbuf.SetLength(0);
}

void CSKKmKanaEngine::ClearRetbuf(void)
{
  retbuf.FillZ();
  retbuf.SetLength(0);
}

void CSKKmKanaEngine::PrepareRetbuf(void)
{
  ClearRetbuf();
  if (kakutei.Length()) {
    retbuf.Copy(kakutei);
  }
  if (regbuf.Length()) {
    retbuf.Append(regbuf);
  }
  retbuf.Append(iBuffer);
  if (post_char) {
    retbuf.Append(post_char);
  }
  retbuf.ZeroTerminate();
}

TPtrC CSKKmKanaEngine::Buffer(void)
{
  return iBuffer.Ptr();
}


TPtrC CSKKmKanaEngine::Retbuf(void)
{
  return retbuf.Ptr();
}

void  CSKKmKanaEngine::show_indicator(void)
{
  if (hankaku) {
    iIndicator->SetState(CFepIndicator::EStateHankaku);
  } else if (katakana) {
    iIndicator->SetState(CFepIndicator::EStateKatakana);
  } else {
    iIndicator->SetState(CFepIndicator::EStateKana);
  }
}

void  CSKKmKanaEngine::set_hiragana(void)
{
  katakana = 0;
  hankaku = 0;
  show_indicator();
}

void  CSKKmKanaEngine::set_katakana(void)
{
  katakana = 1;
  hankaku = 0;
  show_indicator();
}

void  CSKKmKanaEngine::set_hankaku(void)
{
  katakana = 0;
  hankaku = 1;
  show_indicator();
}

void  CSKKmKanaEngine::toggle_hankaku(TUint capabilities)
{
  if (!hankaku) {
    if (capabilities & (TCoeInputCapabilities::EJapaneseKatakanaHalfWidth | TCoeInputCapabilities::EAllText)) {
      set_hankaku();
    }
  } else {
    if (capabilities & (TCoeInputCapabilities::EJapaneseHiragana | TCoeInputCapabilities::EAllText)) {
      set_hiragana();
    } else if (capabilities & TCoeInputCapabilities::EJapaneseKatakanaFullWidth) {
      set_katakana();
    }
  }
}

void  CSKKmKanaEngine::toggle_katakana(TUint capabilities)
{
  if (!katakana) {
    if (capabilities & (TCoeInputCapabilities::EJapaneseKatakanaFullWidth | TCoeInputCapabilities::EAllText)) {
      set_katakana();
    }
  } else {
    if (capabilities & (TCoeInputCapabilities::EJapaneseHiragana | TCoeInputCapabilities::EAllText)) {
      set_hiragana();
    }
  }
}

void  CSKKmKanaEngine::remove_prev_strs(void)
{
  int l = strlen(prev_strs);
  if (l) {
    prev_strs[l-1] = 0;
  }
}
